/********************************************************************
* Description:  hal_mxgpio.c
*               This file, 'hal_mxgpio.c', is a HAL component that 
*               provides a driver for the DMP MX-pluse GPIO.
*
* Author: DMP Electronics INC.
* License: GPL Version 2
*    
* Copyright (c) 2012 All rights reserved.
*
* Last change: 
********************************************************************/

#include "rtapi.h"			/* RTAPI realtime OS API */
#include "rtapi_ctype.h"		/* isspace() */
#include "rtapi_string.h" 		/*strcmp() */
#include "rtapi_math.h"
#include "rtapi_app.h"			/* RTAPI realtime module decls */
//#include <linux/interrupt.h>
#include "hal.h"			/* HAL public API decls */

#define rtapi_inb inb
#define rtapi_outb outb
#include <asm/io.h>

#define GPIO0 0x0078
#define GPIO1 0x0079
#define GPIO3 0x007B

#define DIR1 0x0098

/* module
 information */
MODULE_AUTHOR("DMP Electronics INC.");
MODULE_DESCRIPTION("DMP MX-PLUSE GPIO Driver for EMC HAL");
MODULE_LICENSE("GPL");
static char *cfg = "VP00 0xFF";	/* config string, default GPIO and DIR */
RTAPI_MP_STRING(cfg, "config string");

/***********************************************************************
*                STRUCTURES AND GLOBAL VARIABLES                       *
************************************************************************/

/* this structure contains the runtime data needed by the
   GPIo driver for a single port
*/

typedef struct
{
	unsigned short base_addr;	/* base I/O address (0x78, etc.) */
	unsigned short dir_addr;    /* base I/O dir address */
	unsigned char data_dir[8];	/* non-zero if pins 0..7 are output */
	hal_bit_t *data_in[16];		/* ptrs for input pins 0..7 */
    hal_bit_t *data_out[8];		/* ptrs for output pins 0..7 */
   	hal_bit_t data_inv[8];		/* polarity params for output pins 0..7 */
    hal_bit_t data_reset[8];	/* reset flag for output pins 0..7 */
    long long write_time;
    unsigned char outdata;
    unsigned char reset_mask;	/* reset flag for pin 0..7 */
    unsigned char reset_val;	/* reset values for pin 0..7 */
	hal_u32_t reset_time;
	hal_u32_t debug1, debug2;
} mxgpio_t;

/* pointer to array of mxgpio_t structs in shared memory, 1 per port */
static mxgpio_t *port_data_array;

/* other globals */
static int comp_id;		/* component ID */
static int num_ports;		/* number of ports configured */

static unsigned long ns2tsc_factor;
#define ns2tsc(x) (((x) * (unsigned long long)ns2tsc_factor) >> 12)

/***********************************************************************
*                  LOCAL FUNCTION DECLARATIONS                         *
************************************************************************/

static unsigned char read_sb_regb(unsigned char idx);
static unsigned short read_sb_regw(unsigned char idx);
static unsigned long read_sb_reg(unsigned char idx);
static void write_sb_regb(unsigned char idx, unsigned char val);
static void write_sb_regw(unsigned char idx, unsigned short val);
static void write_sb_reg(unsigned char idx, unsigned long val);

/* These are the functions that actually do the I/O
   everything else is just init code
*/

static void read_port(void *arg, long period);
static void reset_port(void *arg, long period);
static void write_port(void *arg, long period);
static void read_all(void *arg, long period);
static void write_all(void *arg, long period);

/* 'pins_and_params()' does most of the work involved in setting up
   the driver.  It parses the command line (argv[]), then if the
   command line is OK, it calls hal_init(), allocates shared memory
   for the mxgpio_t data structure(s), and exports pins and parameters
   It does not set up functions, since that is handled differently in
   realtime and user space.
*/
static int pins_and_params(char *argv[]);

static void parse_port_addr(char *cp, unsigned short *port, unsigned short *dir);
static int export_port(int portnum, mxgpio_t * addr);
static int export_input_pin(int portnum, int pin, hal_bit_t ** base, int n);
static int export_output_pin(int portnum, int pin, hal_bit_t ** dbase, hal_bit_t * pbase, hal_bit_t * rbase, int n);

/***********************************************************************
*                       INIT AND EXIT CODE                             *
************************************************************************/

#define MAX_PORTS 4

#define MAX_TOK ((MAX_PORTS*2)+3)

int rtapi_app_main(void)
{
    char *cp;
    char *argv[MAX_TOK];
    char name[HAL_NAME_LEN + 1];
    int n, retval;


#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
    // this calculation fits in a 32-bit unsigned 
    // as long as CPUs are under about 6GHz
    ns2tsc_factor = (cpu_khz << 6) / 15625ul;
#else
    ns2tsc_factor = 1ll<<12;
#endif

    /* test for config string */
    if (cfg == 0)
    {
		rtapi_print_msg(RTAPI_MSG_ERR, "MX-PLUSE_GPIO: ERROR: no config string\n");
		return -1;
    }
    rtapi_print ( "config string '%s'\n", cfg );
    /* as a RT module, we don't get a nice argc/argv command line, we only
       get a single string... so we need to tokenize it ourselves */
    /* in addition, it seems that insmod under kernel 2.6 will truncate 
       a string parameter at the first whitespace.  So we allow '_' as
       an alternate token separator. */
    cp = cfg;
    for (n = 0; n < MAX_TOK; n++)
    {
		/* strip leading whitespace */
	    while ((*cp != '\0') && ( isspace(*cp) || ( *cp == '_') ))
			cp++;
		/* mark beginning of token */
		argv[n] = cp;
		/* find end of token */
		while ((*cp != '\0') && !( isspace(*cp) || ( *cp == '_') ))
			cp++;
		/* mark end of this token, prepare to search for next one */
		if (*cp != '\0') 
		{
			*cp = '\0';
			cp++;
		}
    }
    for (n = 0; n < MAX_TOK; n++) 
	{
		/* is token empty? */
		if (argv[n][0] == '\0') 
		{
			/* yes - make pointer NULL */
			argv[n] = NULL;
		}
    }
    /* parse "command line", set up pins and parameters */
    retval = pins_and_params(argv);
    if (retval != 0) 
		return retval;
    /* export functions for each port */
    for (n = 0; n < num_ports; n++) 
	{
		/* make read function name */
		rtapi_snprintf(name, sizeof(name), "mxgpio.%d.read", n);
		/* export read function */
		retval = hal_export_funct(name, read_port, &(port_data_array[n]), 0, 0, comp_id);
		if (retval != 0) 
		{
			rtapi_print_msg(RTAPI_MSG_ERR,
			"MX-PLUSE_GPIO: ERROR: Port %d read funct export failed\n", n);
			hal_exit(comp_id);
			return -1;
		}
		/* make write function name */
		rtapi_snprintf(name, sizeof(name), "mxgpio.%d.write", n);
		/* export write function */
		retval = hal_export_funct(name, write_port, &(port_data_array[n]), 0, 0, comp_id);
		if (retval != 0) 
		{
			rtapi_print_msg(RTAPI_MSG_ERR,
			"MX-PLUSE_GPIO: ERROR: port %d write funct export failed\n", n);
			hal_exit(comp_id);
			return -1;
		}
		/* make reset function name */
		rtapi_snprintf(name, sizeof(name), "mxgpio.%d.reset", n);
		/* export write function */
		retval = hal_export_funct(name, reset_port, &(port_data_array[n]), 0, 0, comp_id);
		if (retval != 0) 
		{
			rtapi_print_msg(RTAPI_MSG_ERR,
			"MX-PLUSE_GPIO: ERROR: port %d reset funct export failed\n", n);
			hal_exit(comp_id);
			return -1;
		}
    }
    /* export functions that read and write all ports */
    retval = hal_export_funct("mxgpio.read-all", read_all, port_data_array, 0, 0, comp_id);
    if (retval != 0) 
	{
		rtapi_print_msg(RTAPI_MSG_ERR,
			"MX-PLUSE_GPIO: ERROR: read all funct export failed\n");
		hal_exit(comp_id);
		return -1;
    }
    retval = hal_export_funct("mxgpio.write-all", write_all, port_data_array, 0, 0, comp_id);
    if (retval != 0) 
	{
		rtapi_print_msg(RTAPI_MSG_ERR,
			"MX-PLUSE_GPIO: ERROR: write all funct export failed\n");
		hal_exit(comp_id);
		return -1;
    }
    rtapi_print_msg(RTAPI_MSG_INFO,
	"MX-PLUSE_GPIO: installed driver for %d ports\n", num_ports);
    hal_ready(comp_id);
    return 0;
}

void rtapi_app_exit(void)
{
    int n;
    for (n = 0; n < num_ports; n++) 
	{
        //set GPIO DIR = in;
		rtapi_outb(0x00, port_data_array[n].dir_addr);
		//set ISA Clock
		write_sb_reg(0xC0, read_sb_reg(0xC0) & 0xFFFDFFFFL);
    }
    hal_exit(comp_id);
}

/***********************************************************************
*                  REALTIME PORT READ AND WRITE FUNCTIONS              *
************************************************************************/

static void read_port(void *arg, long period)
{
    mxgpio_t *port;
    int b;
    unsigned char indata, mask;

    port = arg;
    indata = rtapi_inb(port->base_addr);
    /* split the bits into 16 variables (8 regular, 8 inverted) */
    mask = 0x01;
    for (b = 0; b < 16; b += 2) 
    {
    	if(port->data_dir[b/2] == 0)
		{
			*(port->data_in[b]) = indata & mask;
			*(port->data_in[b + 1]) = !(indata & mask);
		}
		mask <<= 1;
    }
}

static void reset_port(void *arg, long period) 
{
    mxgpio_t *port = arg;
    long long deadline, reset_time_tsc;
    unsigned char outdata = (port->outdata&~port->reset_mask) ^ port->reset_val;
   
    if(port->reset_time > period/4) port->reset_time = period/4;
    reset_time_tsc = ns2tsc(port->reset_time);

    if(outdata != port->outdata) 
    {
        deadline = port->write_time + reset_time_tsc;
        while(rtapi_get_clocks() < deadline) {}
        rtapi_outb(outdata, port->base_addr);
    }
}

static void write_port(void *arg, long period)
{
    mxgpio_t *port;
    int b;
    unsigned char outdata, mask;

    port = arg;

    int reset_mask=0, reset_val=0;
    outdata = 0x00;
    mask = 0x01;
    /* assemble output byte for data port from 8 source variables */
    for (b = 0; b < 8; b++) 
    {
    	if(port->data_dir[b] != 0)
	{
    	    /* get the data, add to output byte */
	    if ((*(port->data_out[b])) && (!port->data_inv[b])) 
		outdata |= mask;
			
	    if ((!*(port->data_out[b])) && (port->data_inv[b]))
		outdata |= mask;
			
	    if (port->data_reset[b]) 
	    {
		reset_mask |= mask;
		if(port->data_inv[b]) reset_val |= mask;
	    }
	}
	mask <<= 1;
    }
    /* write it to the hardware */
    rtapi_outb(outdata, port->base_addr);
    port->write_time = rtapi_get_clocks();
    port->reset_val = reset_val;
    port->reset_mask = reset_mask;
    port->outdata = outdata;
}

void read_all(void *arg, long period)
{
    mxgpio_t *port;
    int n;
    port = arg;
    for (n = 0; n < num_ports; n++) 
    {
	read_port(&(port[n]), period);
    }
}

void write_all(void *arg, long period)
{
    mxgpio_t *port;
    int n;
    port = arg;
    for (n = 0; n < num_ports; n++) 
    {
	write_port(&(port[n]), period);
    }
}

/***********************************************************************
*                   LOCAL FUNCTION DEFINITIONS                         *
************************************************************************/

//mxpluse function

/********************************  INT function  *****************************/
static void enable()
{
	asm("sti");// only work on Vortex86SX/DX/MX/MX+...
}

static void disable()
{
	asm("cli");// only work on Vortex86SX/DX/MX/MX+...
}
/*-------------  end of PCI Configuration Space Access Functions  ------------*/


/******************  PCI Configuration Space Access Functions  ***************/
static void io_outpcidw(unsigned long addr, unsigned long val) 
{
	disable();
	outl(addr & 0xfffffffcL, 0xcf8);
	outl(val, 0xcfc);
	enable();
}

static void io_outpciw(unsigned long addr, unsigned short val) 
{
	int i = (int)(addr & 0x03) * 8;
	disable();
	outl(addr & 0xfffffffcL, 0x0cf8);
	outl((inl(0x0cfc) & ~(0x0000ffffL << i)) | ((unsigned long)val << i), 0x0cfc);
	enable();
}

static void io_outpcib(unsigned long addr, unsigned char val) 
{
	int i = (int)(addr & 0x03) * 8;
	disable();
	outl(addr & 0xfffffffcL, 0x0cf8);
	outl((inl(0x0cfc) & ~(0x000000ffL << i)) | ((unsigned long)val << i), 0x0cfc);
	enable();
}

static unsigned long io_inpcidw(unsigned long addr) 
{
    unsigned long tmp;
	disable();
	outl(addr & 0xfffffffcL, 0x0cf8);
    tmp = inl(0x0cfc);
	enable();
    return tmp;

}

static unsigned short io_inpciw(unsigned long addr) 
{
    unsigned short tmp;
	int i = (int)(addr & 0x03) * 8;
	disable();
	outl(addr & 0xfffffffcL, 0x0cf8);
    tmp = (unsigned short)((inl(0x0cfc) >> i) & 0xffffL);
	enable();
    return tmp;

}

static unsigned char io_inpcib(unsigned long addr) 
{
    unsigned char tmp;
	int i = (int)(addr & 0x03) * 8;
	disable();
	outl(addr & 0xfffffffcL, 0x0cf8);
    tmp = (unsigned char)((inl(0x0cfc) >> i) & 0xffL);
	enable();
    return tmp;
}
/*-------------  end of PCI Configuration Space Access Functions  ------------*/

// read south bridge function register
static unsigned char read_sb_regb(unsigned char idx) 
{
	return io_inpcib(0x80003800L+(unsigned long)idx);
}

static unsigned short read_sb_regw(unsigned char idx) 
{
	return io_inpciw(0x80003800L+(unsigned long)idx);
}

static unsigned long read_sb_reg(unsigned char idx) 
{
	return io_inpcidw(0x80003800L+(unsigned long)idx);
}

// write south bridge function register
static void write_sb_regb(unsigned char idx, unsigned char val) 
{
	io_outpcib(0x80003800L+(unsigned long)idx, val);
}

static void write_sb_regw(unsigned char idx, unsigned short val) 
{
	io_outpciw(0x80003800L+(unsigned long)idx, val);
}

static void write_sb_reg(unsigned char idx, unsigned long val) 
{
	io_outpcidw(0x80003800L+(unsigned long)idx, val);
}
//end mxpluse function

static unsigned short atoi(char *cp)
{
    unsigned short result;

    /* initial value */
    result = 0;
    /* test for leading '0x' */
    if (cp[0] == '0') 
    {
	if ((cp[1] == 'X') || (cp[1] == 'x')) 
	{
	    /* leading '0x', skip it */
	    cp += 2;
	}
    }
    /* ok, now parse digits */
    while (*cp != '\0') 
    {
	/* if char is a hex digit, add it to result */
	if ((*cp >= '0') && (*cp <= '9')) 
	{
	    result <<= 4;
	    result += *cp - '0';
	} else if ((*cp >= 'A') && (*cp <= 'F')) 
	{
	    result <<= 4;
	    result += (*cp - 'A') + 10;
	} else if ((*cp >= 'a') && (*cp <= 'f')) 
	{
	    result <<= 4;
	    result += (*cp - 'a') + 10;
	} else 
	{
	    /* not a valid hex digit */
	    return -1;
	}
	/* next char */
	cp++;
    }

    return result;
}

static int pins_and_params(char *argv[])
{
    unsigned short port_addr[MAX_PORTS];
	unsigned short dir_addr[MAX_PORTS];
    int data_dir[MAX_PORTS];
    int i, n, retval;
    unsigned short dir;

    /* clear port_addr and data_dir arrays */
    for (n = 0; n < MAX_PORTS; n++) 
    {
	port_addr[n] = 0;
	data_dir[n] = 0;
	dir_addr[n] = 0;
    }
    /* parse config string, results in port_addr[] and data_dir[] arrays */
    num_ports = 0;
    n = 0;
    dir = 0;

	//set ISA Clock(high speed)
	write_sb_reg(0xC0, read_sb_reg(0xC0) | 0x20000L);
	
    while ((num_ports < MAX_PORTS) && (argv[n] != 0)) 
    {
	parse_port_addr(argv[n], &port_addr[num_ports], &dir_addr[num_ports]);
	if (port_addr[num_ports] < 0) 
	{
	    rtapi_print_msg(RTAPI_MSG_ERR, "MX-PLUSE_GPIO: ERROR: bad GPIO '%s'\n", argv[n]);
	    return -1;
	}
	n++;
	if (argv[n] != 0) 
	{
	    dir = atoi(argv[n]);
	    data_dir[num_ports] = dir;
	    n++;
	}
	num_ports++;
    }
    /* OK, now we've parsed everything */
    if (num_ports == 0) 
    {
	rtapi_print_msg(RTAPI_MSG_ERR, "MX-PLUSE_GPIO: ERROR: no ports configured\n");
	return -1;
    }
    /* have good config info, connect to the HAL */
    comp_id = hal_init("hal_mxgpio");
    if (comp_id < 0) 
    {
	rtapi_print_msg(RTAPI_MSG_ERR, "MX-PLUSE_GPIO: ERROR: hal_init() failed\n");
	return -1;
    }
    /* allocate shared memory for mxgpio data */
    port_data_array = hal_malloc(num_ports * sizeof(mxgpio_t));
    if (port_data_array == 0) 
    {
	rtapi_print_msg(RTAPI_MSG_ERR, "MX-PLUSE_GPIO: ERROR: hal_malloc() failed\n");
	hal_exit(comp_id);
	return -1;
    }
    /* export all the pins and params for each port */
    dir = 0;
    for (n = 0; n < num_ports; n++) 
    {
	/* config addr and direction */
	port_data_array[n].base_addr = port_addr[n];

	port_data_array[n].dir_addr = dir_addr[n];
	
	dir = data_dir[n];
	for(i = 0 ; i < 8; i++)
	    port_data_array[n].data_dir[i] = (dir >> i) & 0x01;

	rtapi_outb(data_dir[n], port_data_array[n].dir_addr);

	/* export all vars */
	retval = export_port(n, &(port_data_array[n]));
	if (retval != 0) 
	{
	    rtapi_print_msg(RTAPI_MSG_ERR, "MX-PLUSE_GPIO: ERROR: port %d var export failed\n", n);
	    hal_exit(comp_id);
	    return retval;
	}
    }
    return 0;
}

static void parse_port_addr(char *cp, unsigned short *port, unsigned short *dir)
{	
    if(strcmp(cp, "VP00") == 0)
    {
		*port = GPIO0;
		*dir = DIR1 + 0;
    }
    else if(strcmp(cp, "VP01") == 0)
    {
		*port = GPIO1;
		*dir = DIR1 + 1;
    }
    else if(strcmp(cp, "VP03") == 0)
    {
		*port = GPIO3;
		*dir = DIR1 + 3;
    }
    else
	{
		*port = 0;
		*dir = 0;
	}
}

static int export_port(int portnum, mxgpio_t * port)
{
    int retval, msg, i;

    /* This function exports a lot of stuff, which results in a lot of
       logging if msg_level is at INFO or ALL. So we save the current value
       of msg_level and restore it later.  If you actually need to log this
       function's actions, change the second line below */
    msg = rtapi_get_msg_level();
    rtapi_set_msg_level(RTAPI_MSG_WARN);

    retval = 0;
    for(i = 0; i < 8; i++)
    {
	if(port->data_dir[i] == 0)
	{
	    /* declare input pins (data port) */
	    retval += export_input_pin(portnum, i, port->data_in, i);
	}
	else
	{
	    /* declare output pins (data port) */
	    retval += export_output_pin(portnum, i, port->data_out, port->data_inv, port->data_reset, i);
	}
    }
    retval += hal_param_u32_newf(HAL_RW, &port->reset_time, comp_id, "mxgpio.%d.reset-time", portnum);
    retval += hal_param_u32_newf(HAL_RW, &port->debug1, comp_id, "mxgpio.%d.debug1", portnum);
    retval += hal_param_u32_newf(HAL_RW, &port->debug2, comp_id, "mxgpio.%d.debug2", portnum);
    port->write_time = 0;
    /* restore saved message level */
    rtapi_set_msg_level(msg);
    return retval;
}

static int export_input_pin(int portnum, int pin, hal_bit_t ** base, int n)
{
    int retval;

    /* export write only HAL pin for the input bit */
    retval = hal_pin_bit_newf(HAL_OUT, base + (2 * n), comp_id, "mxgpio.%d.pin-%02d-in", portnum, pin);
    if (retval != 0) 
	return retval;
    
    /* export another write only HAL pin for the same bit inverted */
    retval = hal_pin_bit_newf(HAL_OUT, base + (2 * n) + 1, comp_id, "mxgpio.%d.pin-%02d-in-not", portnum, pin);
    return retval;
}

static int export_output_pin(int portnum, int pin, hal_bit_t ** dbase, hal_bit_t * pbase, hal_bit_t * rbase, int n)
{
    int retval;

    /* export read only HAL pin for output data */
    retval = hal_pin_bit_newf(HAL_IN, dbase + n, comp_id, "mxgpio.%d.pin-%02d-out", portnum, pin);
    if (retval != 0) return retval;
	
    /* export parameter for polarity */
    retval = hal_param_bit_newf(HAL_RW, pbase + n, comp_id, "mxgpio.%d.pin-%02d-out-invert", portnum, pin);
    if (retval != 0) return retval;

    /* export parameter for reset */
    if (rbase)
	retval = hal_param_bit_newf(HAL_RW, rbase + n, comp_id, "mxgpio.%d.pin-%02d-out-reset", portnum, pin);
    return retval;
}


