/********************************************************************
* Description:  mxpwmgen.c
*               A HAL component that generates Pulse Width Modulation
*
* Author: DMP Electronics INC.
* License: GPL Version 2
*    
* Copyright (c) 2012 All rights reserved.
*
* Last change: 
********************************************************************/
/** This file, 'pwmgen.c', is a HAL component that generates
    Pulse Width Modulation in hardware.
*/

#include "rtapi.h"		/* RTAPI realtime OS API */
#include "rtapi_app.h"		/* RTAPI realtime module decls */
#include "hal.h"		/* HAL public API decls */
#define rtapi_inb inb
#define rtapi_outb outb
#include <asm/io.h>

#define MAX_CHAN 16

/* module information */
MODULE_AUTHOR("DMP Electronics INC.");
MODULE_DESCRIPTION("PWM Generator for EMC HAL");
MODULE_LICENSE("GPL");
#define MAX_OUTPUT 16
int output_pin[MAX_CHAN] = { -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1};
RTAPI_MP_ARRAY_INT(output_pin, 16, "output pin for up to 16 channels");

/***********************************************************************
*                STRUCTURES AND GLOBAL VARIABLES                       *
************************************************************************/

typedef struct {
    long period;			/* length of PWM period, ns */
    unsigned char output_pin;

    hal_bit_t *enable;		/* pin for enable signal */
    hal_float_t *value;		/* command value */
    hal_float_t *scale;		/* pin: scaling from value to duty cycle */
    hal_float_t *offset;	/* pin: offset: this is added to duty cycle */
    double old_scale;		/* stored scale value */
    double scale_recip;		/* reciprocal value used for scaling */
    hal_float_t *pwm_freq;	/* pin: (max) output frequency in Hz */
    double old_pwm_freq;	/* used to detect changes */
    hal_float_t *min_dc;	/* pin: minimum duty cycle */
    hal_float_t *max_dc;	/* pin: maximum duty cycle */
    hal_float_t *curr_dc;	/* pin: current duty cycle */
} pwmgen_t;

/* ptr to array of pwmgen_t structs in shared memory, 1 per channel */
static pwmgen_t *pwmgen_array;

/* other globals */
static int comp_id;		/* component ID */
static int num_chan;		/* number of pwm generators configured */

/***********************************************************************
*                  LOCAL FUNCTION DECLARATIONS                         *
************************************************************************/

static int export_pwmgen(int num, pwmgen_t * addr);
static void update(void *arg, long period);

/***********************************************************************
*                       INIT AND EXIT CODE                             *
************************************************************************/
//mxpluse io function

/********************************  INT function  *****************************/
static void enable(void)
{
	asm("sti");// only work on Vortex86SX/DX/MX/MX+...
}

static void disable(void)
{
	asm("cli");// only work on Vortex86SX/DX/MX/MX+...
}

/******************  PCI Configuration Space Access Functions  ***************/
static void io_outpcidw(unsigned long addr, unsigned long val) 
{
	disable();
	outl(addr & 0xfffffffcL, 0xcf8);
	outl(val, 0xcfc);
	enable();
}

static unsigned long io_inpcidw(unsigned long addr) 
{
    unsigned long tmp;
	disable();
	outl(addr & 0xfffffffcL, 0x0cf8);
    tmp = inl(0x0cfc);
	enable();
    return tmp;

}

/*-------------  end of PCI Configuration Space Access Functions  ------------*/

// read south bridge function register
static unsigned long read_sb_reg(unsigned char idx) 
{
	return io_inpcidw(0x80003800L+(unsigned long)idx);
}

// write south bridge function register
static void write_sb_reg(unsigned char idx, unsigned long val) 
{
	io_outpcidw(0x80003800L+(unsigned long)idx, val);
}
//end mxpluse io function

//pwm code

#define SB_PWMSETTING_REG   (0xd0)
#define SB_GPIOFUNC_REG     (0xc8)

#define PWM_LREG_BASEADDR   (BaseAddress + 0x0c)
#define PWM_HREG_BASEADDR   (BaseAddress + 0x10)
#define PWM_CTRL_BASEADDR   (BaseAddress + 0x14)

static unsigned BaseAddress = 0xfe00;

int PWM_init(void) {
	int i;
	
	//switch all pins to gpio mode
	write_sb_reg(SB_GPIOFUNC_REG, 0x00L );
	//set base address of hardware PWM generators
	write_sb_reg(SB_PWMSETTING_REG, ((read_sb_reg(SB_PWMSETTING_REG) & 0xff7f0000L) | 0x00800000L) + BaseAddress );
	//PWM base clock = 50MHz
	write_sb_reg(SB_PWMSETTING_REG,(read_sb_reg(SB_PWMSETTING_REG) & 0xfeffffffL) + (1L << 24) );
	//disable IRQ of PWM generators
	write_sb_reg(SB_PWMSETTING_REG, read_sb_reg(SB_PWMSETTING_REG) & 0xfff0ffffL );
	//enable pwm generators; initialize to count mode & set count = 0
	for(i = 0; i < 16; i++)
		outl(0x80000000L, PWM_CTRL_BASEADDR + i*12);
	return 0;
}

#define PWM 1L
#define GPIO 0L
int SwitchPIN(int channel, unsigned long mode) {
	if(mode == PWM)
		write_sb_reg(SB_GPIOFUNC_REG, read_sb_reg(SB_GPIOFUNC_REG) | (1L << channel) );
	else if(mode == GPIO)
		write_sb_reg(SB_GPIOFUNC_REG, read_sb_reg(SB_GPIOFUNC_REG) & ~(1L << channel));
	else {
		return -1;
	}
	return 0;
}

int PWM_close(void) {
	int i;
	//switch all pins to gpio mode
	write_sb_reg(SB_GPIOFUNC_REG, 0x00L );
	//disable PWM generators
	for(i = 0; i < 16; i++)
		outl(0x00000000L, PWM_CTRL_BASEADDR + i*12);	
	return 0;
}

int PWM_set(int channel, unsigned long period, unsigned long duty) {
	if ((period == 0L) || (duty == 0L) || (period <= duty))
	{
		return -1;
	}
	
	//Set period & duty
	outl(duty, PWM_HREG_BASEADDR + channel*12);
	outl(period - duty, PWM_LREG_BASEADDR + channel*12);
	return 0;
}
void PWM_enable(int channel) {	
	outl(inl(PWM_CTRL_BASEADDR + channel*12) | 0xC0000000L, PWM_CTRL_BASEADDR + channel*12);	
}

void PWM_disable(int channel) {	
	outl(inl(PWM_CTRL_BASEADDR + channel*12) & 0x7FFFFFFFL, PWM_CTRL_BASEADDR + channel*12);	
}

//end pwm code

int rtapi_app_main(void)
{
    int n, retval;

    for (n = 0; n < MAX_CHAN && output_pin[n] != -1 ; n++) {
		if ((output_pin[n] > MAX_OUTPUT) || (output_pin[n] < 0)) {
			rtapi_print_msg(RTAPI_MSG_ERR,
			"PWMGEN: ERROR: bad output pin '%i', channel %i\n",
			output_pin[n], n);
			return -1;
		}
		//init pwm pin
		PWM_init();
		SwitchPIN(output_pin[n], PWM);
		num_chan++;	
    }
    if (num_chan == 0) {
		rtapi_print_msg(RTAPI_MSG_ERR,
			"PWMGEN: ERROR: no channels configured\n");
		return -1;
    }
	
    /* have good config info, connect to the HAL */
    comp_id = hal_init("mxpwmgen");
    if (comp_id < 0) {
		rtapi_print_msg(RTAPI_MSG_ERR, "PWMGEN: ERROR: hal_init() failed\n");
		return -1;
    }
    /* allocate shared memory for generator data */
    pwmgen_array = hal_malloc(num_chan * sizeof(pwmgen_t));
    if (pwmgen_array == 0) {
		rtapi_print_msg(RTAPI_MSG_ERR,
			"PWMGEN: ERROR: hal_malloc() failed\n");
		hal_exit(comp_id);
		return -1;
    }
    /* export all the variables for each PWM generator */
    for (n = 0; n < num_chan; n++) {
		/* export all vars */
		retval = export_pwmgen(n, &(pwmgen_array[n]));
		if (retval != 0) {
			rtapi_print_msg(RTAPI_MSG_ERR,
			"PWMGEN: ERROR: pwmgen %d var export failed\n", n);
			hal_exit(comp_id);
			return -1;
		}
	    if(rtapi_request_region((BaseAddress + 0x0c) + n*12, 12, hal_comp_name(comp_id)) == NULL) {
			rtapi_print_msg(RTAPI_MSG_ERR,
			"PWMGEN: ERROR: pwmgen %d rtapi request region fail\n", n);
			hal_exit(comp_id);
			return -1;	    
		}
    }
    /* export functions */
    retval = hal_export_funct("pwmgen.update", update,
	pwmgen_array, 1, 0, comp_id);
    if (retval != 0) {
		rtapi_print_msg(RTAPI_MSG_ERR,
			"PWMGEN: ERROR: update funct export failed\n");
		hal_exit(comp_id);
		return -1;
    }
    rtapi_print_msg(RTAPI_MSG_INFO,
	"PWMGEN: installed %d PWM generators\n", num_chan);
    hal_ready(comp_id);
    return 0;
}

void rtapi_app_exit(void)
{
	int n;
	//swatch pwm->gpio
    for (n = 0; n < num_chan ; n++) {
		PWM_disable(output_pin[n]);
		rtapi_release_region((BaseAddress + 0x0c) + n*12, 12);
		//swith pwm->gpio
		SwitchPIN(output_pin[n], GPIO);
		//close pwm pin
		PWM_close();
    }
    hal_exit(comp_id);
}

/***********************************************************************
*              REALTIME STEP PULSE GENERATION FUNCTIONS                *
************************************************************************/

static void update(void *arg, long period)
{
    pwmgen_t *pwmgen;
    int n, freq, duty;
	double tmpdc, outdc;

    /* update the PWM generators */
    pwmgen = arg;
    for (n = 0; n < num_chan; n++) {
	
	/* validate duty cycle limits, both limits must be between
	   0.0 and 1.0 (inclusive) and max must be greater then min */
	if ( *(pwmgen->max_dc) > 1.0 ) {
	    *(pwmgen->max_dc) = 1.0;
	}
	if ( *(pwmgen->min_dc) > *(pwmgen->max_dc) ) {
	    *(pwmgen->min_dc) = *(pwmgen->max_dc);
	}
	if ( *(pwmgen->min_dc) < 0.0 ) {
	    *(pwmgen->min_dc) = 0.0;
	}
	if ( *(pwmgen->max_dc) < *(pwmgen->min_dc) ) {
	    *(pwmgen->max_dc) = *(pwmgen->min_dc);
	}
	/* do scale calcs only when scale changes */
	if ( *(pwmgen->scale) != pwmgen->old_scale ) {
	    /* get ready to detect future scale changes */
	    pwmgen->old_scale = *(pwmgen->scale);
	    /* validate the new scale value */
	    if ((*(pwmgen->scale) < 1e-20)
		&& (*(pwmgen->scale) > -1e-20)) {
		/* value too small, divide by zero is a bad thing */
		*(pwmgen->scale) = 1.0;
	    }
	    /* we will need the reciprocal */
	    pwmgen->scale_recip = 1.0 / *(pwmgen->scale);
	}
	
	if(*(pwmgen->pwm_freq) != pwmgen->old_pwm_freq) {
	    freq = (int)(*(pwmgen->pwm_freq));
	    pwmgen->period = 50000000L / freq;//pwm resolution 20ns
	    /* save freq to detect changes */
	    pwmgen->old_pwm_freq = *(pwmgen->pwm_freq);	    			
	}
	/* convert value command to duty cycle */
	tmpdc = *(pwmgen->value) * pwmgen->scale_recip + *(pwmgen->offset);
	/* limit the duty cycle */
	if (tmpdc >= 0.0) {
	    if ( tmpdc > *(pwmgen->max_dc) ) {
		tmpdc = *(pwmgen->max_dc);
	    } else if ( tmpdc < *(pwmgen->min_dc) ) {
		tmpdc = *(pwmgen->min_dc);
	    }
	    outdc = tmpdc;
	}
	else {
	    outdc = 0.0;
	}
	/* save rounded value to curr_dc param */
	*(pwmgen->curr_dc) = outdc;
	duty = outdc * pwmgen->period;
	PWM_set(pwmgen->output_pin, (unsigned long)pwmgen->period, (unsigned long)duty);//SetPWM	
	if ( *(pwmgen->enable) == 1 ) {
	    PWM_enable(pwmgen->output_pin);
	} else {
	    PWM_disable(pwmgen->output_pin);
	}
	/* move on to next channel */
	pwmgen++;
    }
    /* done */
}


/***********************************************************************
*                   LOCAL FUNCTION DEFINITIONS                         *
************************************************************************/

static int export_pwmgen(int num, pwmgen_t * addr)
{
    int retval, msg;

    /* This function exports a lot of stuff, which results in a lot of
       logging if msg_level is at INFO or ALL. So we save the current value
       of msg_level and restore it later.  If you actually need to log this
       function's actions, change the second line below */
    msg = rtapi_get_msg_level();
    rtapi_set_msg_level(RTAPI_MSG_WARN);

    /* export paramameters */
    retval = hal_pin_float_newf(HAL_IO, &(addr->scale), comp_id,
	    "pwmgen.%d.scale", num);
    if (retval != 0) {
	return retval;
    }
    retval = hal_pin_float_newf(HAL_IO, &(addr->offset), comp_id,
	    "pwmgen.%d.offset", num);
    if (retval != 0) {
	return retval;
    }
    retval = hal_pin_float_newf(HAL_IO, &(addr->pwm_freq), comp_id,
	    "pwmgen.%d.pwm-freq", num);
    if (retval != 0) {
	return retval;
    }
    retval = hal_pin_float_newf(HAL_IO, &(addr->min_dc), comp_id,
	    "pwmgen.%d.min-dc", num);
    if (retval != 0) {
	return retval;
    }
    retval = hal_pin_float_newf(HAL_IO, &(addr->max_dc), comp_id,
	    "pwmgen.%d.max-dc", num);
    if (retval != 0) {
	return retval;
    }
    retval = hal_pin_float_newf(HAL_OUT, &(addr->curr_dc), comp_id,
	    "pwmgen.%d.curr-dc", num);
    if (retval != 0) {
	return retval;
    }
    /* export pins */
    retval = hal_pin_bit_newf(HAL_IN, &(addr->enable), comp_id,
	    "pwmgen.%d.enable", num);
    if (retval != 0) {
	return retval;
    }
    *(addr->enable) = 1;
    retval = hal_pin_float_newf(HAL_IN, &(addr->value), comp_id,
	    "pwmgen.%d.value", num);
    if (retval != 0) {
	return retval;
    }
	*(addr->value) = 0.0;
    addr->output_pin = output_pin[num];
    /* set default pin values */
    *(addr->scale) = 1.0;
    *(addr->offset) = 0.0;
    *(addr->pwm_freq) = 0;
    *(addr->min_dc) = 0.0;
    *(addr->max_dc) = 1.0;
    *(addr->curr_dc) = 0.0;
    /* init other fields */
    addr->period = 50000;
    addr->old_scale = *(addr->scale) + 1.0;
    addr->old_pwm_freq = -1;
    /* restore saved message level */
    rtapi_set_msg_level(msg);
    return 0;
}
